import urllib.parse
from dataclasses import dataclass
from datetime import timedelta
from functools import cached_property

import dateutil.parser
import downloader
import pandas as pd
import pywebcopy
import requests
from bs4 import BeautifulSoup
from plib import Path

from ..asset_types import Due, SaveItem
from . import base


@dataclass
class Url(SaveItem):
    url: str
    root_url: str
    folder: Path

    @property
    def relative_url(self):
        url = self.url.replace(self.root_url, "")
        url = urllib.parse.urlparse(url).path
        return url

    @property
    def local_file(self):
        return self.folder.subpath(*self.relative_url.split("/"))

    @property
    def is_html_content(self):
        is_html = self.local_file.suffix == ".html"
        if not is_html:
            headers = requests.get(self.url, headers=self.headers, stream=True).headers
            content_type = headers.get("content-type") or headers.get("Content-Type")
            is_html = "text/html" in content_type
        return is_html

    def save(self):
        dest = self.local_file
        if self.is_html_content:
            base_tag = f"<base href='{self.root_url}'>"
            dest = dest.with_suffix(".html")
            dest.text = base_tag + requests.get(self.url).text
        else:
            downloader.download(self.url, dest)
        dest.tag = 0
        dest.unpack_if_archive()

    @property
    def display_title(self):
        return self.relative_url

    @property
    def save_id(self):
        return self.relative_url

    @property
    def headers(self):
        browser = (
            "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/85.0.4183.83 Safari/537.36"
        )
        return {"User-Agent": browser}


class Checker(base.Checker):
    def make_item(self, item):
        return (
            Due(*item)
            if isinstance(item, tuple)
            else Url(item, self.url, self.path.parent / "Homepage")
        )

    def get_urls(self):
        return self.course.websites

    @cached_property
    def url(self):
        urls = self.get_urls()
        root_url = next(iter(urls))
        if not root_url.endswith("/"):
            root_url = f"{root_url}/"
        return root_url

    def sub_urls(self):
        return next(iter(self.get_urls().values()))

    @property
    def path(self):
        return super().path / "Homepage.html"

    def get_items(self):
        for sub_url in self.sub_urls():
            url = urllib.parse.urljoin(self.url, sub_url).strip("/")
            yield from self.get_url_items(url)

    def get_url_items(self, url):
        html_content: bytes = requests.get(url).content
        extractors = (self.get_content_items,)  # get_calendar_items
        for extractor in extractors:
            yield from extractor(html_content)  # noqa

    def get_content_items(self, html_content):
        soup = BeautifulSoup(html_content, features="lxml")
        for link in soup.find_all("a"):
            yield urllib.parse.urljoin(self.url, link.get("href"))

    def get_calendar_items(self, html_content: bytes):
        with Path.tempfile() as tmp:
            tmp.byte_content = html_content
            tables = pd.read_html(tmp)

        due_keywords = ("Due", "Assignments")
        for table in tables:
            table.reset_index(drop=True, inplace=True)

            for due_keyword in due_keywords:
                if due_keyword in table.columns:
                    calendar_items = table[~table[due_keyword].isnull()]
                    for _, item in calendar_items.iterrows():
                        date_key = "Date" if "Date" in table.columns else due_keyword
                        name_key = (
                            "Assignment"
                            if "Assignment" in table.columns
                            else due_keyword
                        )
                        date = item[date_key]
                        if date_key == "Due":
                            date = " ".join(date.split(" ")[1:])
                        if date:
                            date = date.replace("Tues", "Tue")
                            parsed_date = dateutil.parser.parse(date)
                            due_time = parsed_date + timedelta(hours=23, minutes=30)
                            parsed_name = self.course.assignment_name(item[name_key])
                            yield parsed_name, due_time

    def should_check(self) -> bool:
        return bool(self.course.websites)

    def check_new_items(self):
        self.save_website()
        super().check_new_items()

    def save_website(self, offline=False):
        if offline:
            self.save_offline()
        else:
            self.save_online()

    def save_offline(self):
        folder = self.path.with_suffix("")
        pywebcopy.save_webpage(self.url, str(folder.parent), project_name=folder.name)
        index_path = next(folder.rglob("index.html"))
        index_path_symlink = self.path
        index_path_symlink.symlink_to(index_path)

    def save_online(self):
        if not self.path.exists():
            self.path.text = f"<script>location.href = '{self.url}'</script>"
            self.path.tag = 9999



from json import JSONDecodeError

import requests
from canvasapi import Canvas, exceptions
from gcsa.google_calendar import GoogleCalendar
from oauthlib.oauth2.rfc6749.errors import InvalidClientError

from .path import Path


def is_valid_account(url, key):
    return is_valid_url(url, key) and is_valid_key(url, key)


def is_valid_url(url, key):
    invalid_exceptions = (
        requests.exceptions.MissingSchema,
        requests.exceptions.JSONDecodeError,
    )
    ignored_exceptions = (
        exceptions.InvalidAccessToken,
        StopIteration,
        exceptions.ResourceDoesNotExist,
    )
    is_valid = usage_works(url, key, invalid_exceptions, ignored_exceptions)
    return is_valid


def is_valid_key(url, key):
    invalid_exceptions = (
        exceptions.InvalidAccessToken,
        exceptions.ResourceDoesNotExist,
    )
    ignored_exceptions = (StopIteration,)
    return usage_works(url, key, invalid_exceptions, ignored_exceptions)


def usage_works(url, key, failure_exceptions, ignored_exceptions):
    works = is_valid_url_scheme(url)
    if works:
        client = Canvas(url, key)
        accounts = client.get_accounts()
        try:
            next(iter(accounts))
        except failure_exceptions:
            works = False
        except ignored_exceptions:
            pass
    return works


def is_valid_url_scheme(url):
    is_valid = True
    try:
        api_url = f"{url}/api/v1"
        requests.head(api_url)
    except ValueError:
        is_valid = False
    return is_valid


def google_calendar_credentials_valid(calendar_id=None):
    is_valid = True
    failure_exceptions = (
        FileNotFoundError,
        JSONDecodeError,
        ValueError,
        InvalidClientError,
    )
    args = (calendar_id,) if calendar_id else ()
    try:
        GoogleCalendar(*args, credentials_path=Path.calendar_credentials)
    except failure_exceptions:
        is_valid = False
    return is_valid



from abc import ABC
from dataclasses import dataclass

from canvasdl import client

from .. import base


@dataclass
class Checker(base.Checker, ABC):
    @property
    def api(self):
        return client.get_course(self.course.id)



from dataclasses import dataclass
from typing import Dict

from ..utils import Path, time
from . import SaveItem


@dataclass
class Announcement(SaveItem):
    id: int
    user_id: int
    course_id: int
    editor_id: int | None
    accepted_id: int | None
    duplicate_id: int | None
    number: int
    type: str
    title: str
    content: str
    document: str
    category: str
    subcategory: str
    subsubcategory: str
    flag_count: int
    star_count: int
    view_count: int
    unique_view_count: int
    vote_count: int
    reply_count: int
    unresolved_count: int
    is_locked: bool
    is_pinned: bool
    is_private: bool
    is_endorsed: bool
    is_answered: bool
    is_student_answered: bool
    is_staff_answered: bool
    is_archived: bool
    is_anonymous: bool
    is_megathread: bool
    anonymous_comments: bool
    approved_status: str
    created_at: str
    updated_at: str
    deleted_at: None
    pinned_at: str | None
    anonymous_id: int
    vote: int
    is_seen: bool
    is_starred: bool
    is_watched: None
    glanced_at: str | None
    new_reply_count: int
    duplicate_title: None
    user: Dict

    save_folder: Path = None

    @property
    def short_title(self):
        return self.title

    @property
    def mtime(self):
        return self.created_at

    def save(self):
        save_path = self.save_folder / f"edstem_{self.save_id}.yaml"
        save_path.yaml = {
            "title": self.short_title,
            "created_at": self.mtime,
            "message": self.content,
        }
        save_path.mtime = time.parse_time(self.mtime)

    @property
    def display_title(self):
        return f"{self.short_title} ({time.export_time(self.mtime)})"

    @property
    def save_id(self):
        return self.id



import plib

root = plib.Path(__file__).parent.parent


class Path(plib.Path):
    templates = root / "assets" / "templates"
    announ_css = templates / "announ.css"
    assets: plib.Path = plib.Path.assets / root.name
    content_assets = assets / "content"
    config = assets / "config.yaml"
    calendar_credentials = assets / "credentials.json"
    calendar_token = assets / "token.pickle"
    courses = assets / "courses" / "courses.yaml"

    school = plib.Path.docs / "School"

    @classmethod
    def content_path(cls, course, names=None):
        names = names or ()
        path = Path.content_assets.subpath(*names, course).with_suffix(".yaml")
        path = Path(path)  # convert to right type
        return path



import sys
import threading

WITH_UI = True

try:
    from PyQt6 import QtWidgets  # noqa: autoimport
except (ModuleNotFoundError, ImportError):
    WITH_UI = False

cond = threading.Condition()


class UserInterface:
    tasks = []
    to_show = 0
    checked = 0

    @classmethod
    def tasks_are_remaining(cls, done, to_check):
        return WITH_UI & (done < cls.to_show or cls.checked < to_check)

    @classmethod
    def show_progress(cls, to_check):
        done = 0
        app = None

        while cls.tasks_are_remaining(done, to_check):
            if not cls.tasks or cls.checked < to_check:
                with cond:
                    cond.wait()
            if cls.tasks:
                if app is None:
                    app = QtWidgets.QApplication(sys.argv)
                download_progress = UserInterface.tasks.pop(0)
                download_progress.grant_widget()
                done += 1
                with cond:  # allow new tasks
                    cond.notify_all()

    @classmethod
    def add_check(cls):
        cls.checked += 1
        with cond:
            cond.notify_all()

    @classmethod
    def request_widget(cls, download_progress):
        if download_progress not in cls.tasks:
            with cond:
                cls.to_show += 1
                cls.tasks.append(download_progress)
                cond.notify_all()



import os

from PyQt6 import QtWidgets
from PyQt6.QtCore import Qt

# disable annoying warnings
os.environ["QT_LOGGING_RULES"] = "*.debug=false;qt.qpa.*=false"

MARGIN = 10
TEXT_HEIGHT = 174
HEIGHT = 250
WIDTH = 370
PROGRESS_HEIGHT = 14


class WidgetUI(QtWidgets.QDialog):
    def __init__(self, message=""):
        super().__init__()
        self.initUI(message)
        self.onCancel = None
        self.onOpen = None
        self.onClose = None

    def initUI(self, message):
        scroll = QtWidgets.QScrollArea(self)
        scroll.move(MARGIN, MARGIN)
        scroll.setFixedWidth(WIDTH - 2 * MARGIN)
        scroll.setFixedHeight(TEXT_HEIGHT)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        scroll.setWidgetResizable(True)

        vbox = QtWidgets.QVBoxLayout(scroll)
        for m in message.split("\n"):
            vbox.addWidget(QtWidgets.QLabel(m))

        scrollwidget = QtWidgets.QWidget(self)
        scrollwidget.setLayout(vbox)
        scroll.setWidget(scrollwidget)

        progress = QtWidgets.QProgressBar(self)
        progress.move(MARGIN, TEXT_HEIGHT + 2 * MARGIN)
        progress.setFixedWidth(WIDTH - 2 * MARGIN)
        progress.setFixedHeight(PROGRESS_HEIGHT)

        cancel = QtWidgets.QPushButton("Cancel", self)
        cancel.move(MARGIN, TEXT_HEIGHT + 3 * MARGIN + PROGRESS_HEIGHT)

        open = QtWidgets.QPushButton("Open after download", self)
        open.move(MARGIN + 100, TEXT_HEIGHT + 3 * MARGIN + PROGRESS_HEIGHT)

        close = QtWidgets.QPushButton("OK", self)
        close.move(MARGIN + 250, TEXT_HEIGHT + 3 * MARGIN + PROGRESS_HEIGHT)

        self.setWindowTitle("School")
        self.setWindowFlag(Qt.WindowType.FramelessWindowHint)
        self.setWindowFlag(Qt.WindowType.WindowStaysOnTopHint)

        self.focusWidget()

        self.setGeometry(1500, 580, WIDTH, HEIGHT)
        self.show()

        self.scroll = scroll
        self.vbox = vbox
        self.scrollwidget = scrollwidget
        self.progress = progress

        self.cancel = cancel
        self.open = open
        self.closeButton = close



import cli
from libs.shortcutmaker import ShortcutMaker
from rich.prompt import Confirm, Prompt

from canvasdl import client
from canvasdl.asset_types import Course
from canvasdl.checkers.canvas import videos
from canvasdl.utils import Path, config


def make_courses():
    courses = client.get_courses()
    courses = ask_courses(courses)
    courses = save_courses(courses)
    return courses


def save_courses(courses: list[Course]):
    make_shortcuts(courses)
    for course in courses:
        (Path.school / course.name).mkdir(parents=True, exist_ok=True)

    courses_dict = [c.dict() for c in courses]
    if config.update_content:
        Path.courses.yaml = courses_dict

    return courses_dict


# no longer used
def parse_courses(courses):
    def parse_name(name: str):
        delim = " - "
        return name.split(delim)[1] if delim in name else name

    def include_course(course):
        return hasattr(course, "original_name")  # check if nickname given

    courses = [c for c in courses if include_course(c)]
    parsed_courses = []

    for c in courses:
        course = Course(name=parse_name(c.name), id=c.id)
        video_id = videos.Checker(course).get_folder_id(c.original_name)
        course.video_id = video_id
        parsed_courses.append(course)

    courses = sorted(parsed_courses, key=lambda item: item.name.lower())
    return courses


def make_shortcuts(courses: list[Course]):
    try:
        _make_shortcuts(courses)
    except (ImportError, FileNotFoundError):
        # disable on non-personal machines
        pass


def _make_shortcuts(courses: list[Course]):
    shortcut_maker = ShortcutMaker()
    for i, course in enumerate(courses):
        shortcut_name = course.name.replace(" ", "_").replace("'", "")
        url = f"{config.API_URL}/courses/{course.id}"

        shortcuts = {
            f"checkpoint {shortcut_name}": f'control "c:{10 + i}"',
            # 10, 11, 12 -> 1, 2, 3, ..
            f"checkpoint choose {shortcut_name}": f'control shift "c:{10 + i}"',
            # 67, 68, 69 -> F1, F2, F3, ..
            f"xdg-open {url}": f'control "c:{67 + i}"',
            "": "",
        }

        for target, hotkey in shortcuts.items():
            shortcut_maker.add_shortcut(hotkey, target)

    shortcut_maker.save_shortcuts()


def ask_courses(courses):
    synced_courses = []

    for course in courses:
        name = get_course_name(course)
        question = f"Do you want to synchronize {name}?"
        do_sync = Confirm().ask(question, default=True)
        if do_sync:
            course = Course(name=name, id=course.id)
            original_name = (
                course.original_name
                if hasattr(course, "original_name")
                else course.name
            )

            question = "Do you have an rclone source you want to synchronize with?"
            if Confirm().ask(question):
                rclone_name = Prompt().ask("Rclone source name")
                course.drive_name = rclone_name

            question = "Do you want to synchronize the Piazza posts in this course?"
            if Confirm().ask(question):
                url = Prompt().ask("Piazza url for the course")
                course.piazza_id = url.split("/")[-1]

            question = "Does the course have a custom website you want to synchronize?"
            if Confirm().ask(question):
                root_url = Prompt().ask("Course webpage root url")
                question = (
                    "Does the course have any additional specific subpages you want to"
                    " synchronize?"
                )
                sub_pages = []
                while Confirm().ask(question):
                    sub_pages.append(Prompt.ask("Course subpage sub url"))
                if not sub_pages:
                    sub_pages = [""]
                course.websites = {root_url: sub_pages}

            with cli.status("Saving course"):
                course.video_id = videos.Checker(course).get_folder_id(original_name)
            synced_courses.append(course)

    return synced_courses


def get_course_name(course_info):
    delim = " - "
    name = course_info.name
    if delim in name:
        name = name.split(delim)[1]
    return name



import inspect
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

from canvasdl.asset_types import Course, Section
from canvasdl.ui.progressmanager import ProgressManager
from canvasdl.ui.userinterface import UserInterface
from canvasdl.utils import Path, config

from ..asset_types import SaveItem
from ..ui.downloadprogress import DownloadProgress


@dataclass
class Checker:
    course: Course
    old_content: Any = None

    @property
    def path(self):
        return Path.school / self.course.name

    @classmethod
    def names(cls) -> Iterable[str]:
        root_path = Path(__file__).parent
        full_path = Path(inspect.getfile(cls))
        relative_path = full_path.relative_to(root_path)
        return relative_path.with_suffix("").parts

    @classmethod
    def make_item(cls, item):
        return SaveItem()

    def get_items(self) -> list[Any]:
        raise NotImplementedError

    def __post_init__(self):
        ProgressManager.progress.increase_amount()
        self.content_path = Path.content_path(self.course.name, self.names())

    def load_saved_content(self):
        self.old_content = self.content_path.yaml

    def should_check(self):
        return True

    def check(self):
        self.check_new_content()
        ProgressManager.progress.increase_progress()
        UserInterface.add_check()

    def check_new_content(self):
        self.load_saved_content()
        self.check_new_items()
        self.after_check()

    def check_new_items(self):
        items = [self.make_item(it) for it in self.get_items()]
        items = self.get_new_items(items)
        if items:
            self.process_new_items(items)

    def get_new_items(self, items):
        item_mapper = {it.save_id: it for it in items if self.is_new(it)}
        unique_items = list(item_mapper.values())
        return unique_items

    def is_new(self, item: SaveItem):
        return item.save_id not in self.old_content

    def process_new_items(self, items: list[SaveItem]):
        download_items = [it for it in items if it.should_download()]
        if download_items:
            self.download_new_items(download_items)

        if config.save_content:
            self.save_new_content(items)

    def download_new_items(self, items: list[SaveItem]):
        section = Section(self.path, self.course, items)
        download_progress = DownloadProgress(section)
        for item in items:
            item.save()

        self.export_downloads()
        download_progress.enable_show()

    def save_new_content(self, items: list[SaveItem]):
        self.old_content = (self.old_content or {}) | {
            item.save_id: "" for item in items
        }
        self.content_path.yaml = self.old_content

    def export_downloads(self):
        pass

    def after_check(self):
        pass



import argparse
import sys


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--configure",
        default=False,
        const=True,
        help="Create new config file",
        action="store_const",
    )
    # don't parse arguments if invoked with pytest
    parser_args = ([],) if "pytest" in sys.modules else ()
    args = parser.parse_args(*parser_args)
    return args



from collections import defaultdict

from canvasdl.asset_types import File, Section

from ...ui.downloadprogress import DownloadProgress
from ...utils import config
from . import base


class Checker(base.Checker):
    def get_items(self) -> list[File]:
        files = self.api.get_files(sort="updated_at", order="desc")  # most recent first

        new_files = []
        for file in files:
            if file.updated_at != self.old_content.get(file.id, ""):
                new_files.append(file)
            else:
                break  # next files are less recent so no need to check

        return new_files

    def make_item(self, item):
        item = File.from_response(item)
        item.root = self.path
        return item

    def is_new(self, item: File):
        return True

    def process_new_items(self, items: list[File]):
        file_sections = defaultdict(lambda: [])
        for item in items:
            key = item.get_full_path().parent
            file_sections[key].append(item)

        for path, section_items in file_sections.items():
            section = Section(path, self.course, section_items)
            download_progress = DownloadProgress(section)
            for item in section_items:
                item.save()
            download_progress.enable_show()

        if config.save_content:
            self.save_new_content(items)

    def save_new_content(self, items: list[File]):
        self.old_content |= {item.id: item.updated_at for item in items}
        self.content_path.yaml = self.old_content



from dataclasses import dataclass
from datetime import datetime, timedelta

from canvasdl import client
from canvasdl.asset_types import Announcement

from .. import announ
from . import base


@dataclass
class Checker(base.Checker, announ.Checker):
    check_time: str = ""

    @classmethod
    def current_time() -> str:
        return (datetime.now() - timedelta(days=365)).isoformat()

    def load_saved_content(self):
        super().load_saved_content()

    def make_item(self, item):
        item = Announcement.from_response(item)
        item.save_folder = self.save_folder
        return item

    def get_items(self):
        last_check_time: str = self.content_check_path.text or self.current_time()
        self.check_time = datetime.now().isoformat()
        announcements = client.get_announcements(
            [self.api], start_date=last_check_time, end_date=self.check_time
        )
        return announcements

    @property
    def content_check_path(self):
        return self.content_path.with_suffix(".txt")

    def after_check(self):
        self.content_check_path.text = self.check_time



from canvasdl.ui.progressmanager import ProgressManager


def main():
    with ProgressManager():
        from . import starter  # noqa: autoimport

        starter.check_changes()


if __name__ == "__main__":
    main()



from pathlib import Path

from setuptools import find_packages, setup

NAME = "canvasdl"
version = "1.4.6.post4"


def read(filename):
    try:
        with open(filename) as fp:
            content = fp.readlines()
    except FileNotFoundError:
        content = []
    return content


setup(
    author="Quinten Roets",
    author_email="quinten.roets@gmail.com",
    url="https://github.com/quintenroets/canvasdl",
    download_url=(
        f"https://github.com/quintenroets/canvasdl/archive/refs/tags/v{version}.tar.gz"
    ),
    description="course content synchronizer",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    name=NAME,
    version=version,
    packages=find_packages(),
    setup_requires=read("setup_requirements.txt"),
    install_requires=read("requirements.txt"),
    package_data={"canvasdl": ["assets/**"]},
    entry_points={
        "console_scripts": [
            "canvasdl = canvasdl:main",
            "canvasdlcheck = canvasdl.core.checkservice:main",
        ]
    },
)



from dataclasses import dataclass
from typing import Any

from ..utils import Path, time
from . import base
from .streams import Streams


@dataclass
class ViewUrls:
    ViewerUrl: str
    EmbedUrl: str | None
    ShareSettingsUrl: str | None
    DownloadUrl: str | None
    CaptionDownloadUrl: str | None
    EditorUrl: str | None
    ThumbnailUrl: str


@dataclass
class Video(base.SaveItem):
    Description: str | None
    StartTime: str
    Duration: float
    MostRecentViewPosition: float
    CreatedBy: Any
    Urls: ViewUrls
    Folder: str
    FolderDetails: Any
    PercentCompleted: int | None
    Id: str
    Name: str

    save_folder: Path | None = None

    @property
    def display_title(self):
        return self.Name

    def __post_init__(self):
        delim = " - "
        if delim in self.Name:
            self.Name = self.Name.split(delim)[1]

    @property
    def mtime(self):
        return time.parse_time(self.StartTime)

    @property
    def save_id(self):
        return self.Id

    def save(self):
        save_path = self.save_folder / f"{self.Id}.yaml"
        saved_video = SavedVideo(
            title=self.Name,
            mtime=self.mtime,
            url=self.Urls.ViewerUrl,
            duration=self.Duration,
        )
        save_path.yaml = saved_video.dict()
        save_path.mtime = self.mtime


@dataclass
class SavedVideo(base.Item):
    title: str
    mtime: float
    url: str
    duration: float
    local_url: Path | None = None

    @property
    def id(self):
        return self.url.split("?id=")[1]

    @property
    def filename_title(self):
        return self.title.replace("/", "_")

    def download(self, info: dict, folder: Path):
        saved_info = dict(
            folder=folder / self.id, name=self.filename_title, mtime=self.mtime
        )
        streams_info = {"streams": info["Delivery"]["Streams"]} | saved_info
        streams = Streams.from_dict(streams_info)
        streams.download()
